from functools import wraps

from django.utils.decorators import available_attrs

from .utils import get_throttling_validators


def add_attribute_wrapper(name, value):

    def decorator(view_func):
        def _wrapper(*args, **kwargs):
            return view_func(*args, **kwargs)
        setattr(_wrapper, name, value)
        return wraps(view_func, assigned=available_attrs(view_func))(_wrapper)

    return decorator


def throttling(*validators, keep_default=True):
    """
    Adds throttling validators to a function.
    """

    throttling_validators = list(validators)
    if keep_default:
        throttling_validators += list(get_throttling_validators('default_validators'))
    return add_attribute_wrapper('throttling_validators', throttling_validators)


def throttling_all(*validators, keep_default=True):
    """
    Adds throttling validators to a class.
    """
    def decorator(klass):
        dispatch = getattr(klass, 'dispatch')
        setattr(klass, 'dispatch', throttling(*validators, keep_default=keep_default)(dispatch))
        return klass
    return decorator


def throttling_exempt():
    """
    Marks a function as being exempt from the throttling protection.
    """
    return add_attribute_wrapper('throttling_validators', ())


def throttling_exempt_all(klass):
    """
    Marks a class as being exempt from the throttling protection.
    """
    dispatch = getattr(klass, 'dispatch')
    setattr(klass, 'dispatch', throttling_exempt()(dispatch))
    return klass


def hide_request_body():
    """
    Marks a function as being exempt from storing request base to DB.
    """
    return add_attribute_wrapper('hide_request_body', True)


def hide_request_body_all(klass):
    """
    Marks a class as being exempt from storing request base to DB.
    """
    dispatch = getattr(klass, 'dispatch')
    setattr(klass, 'dispatch', hide_request_body()(dispatch))
    return klass


def log_exempt():
    """
    Marks a function as being exempt from whole log.
    """
    return add_attribute_wrapper('log_exempt', True)


def log_exempt_all(klass):
    """
    Marks a class as being exempt from whole log.
    """
    dispatch = getattr(klass, 'dispatch')
    setattr(klass, 'dispatch', log_exempt()(dispatch))
    return klass



from .throttling import PerRequestThrottlingValidator


default_validators = (
    PerRequestThrottlingValidator(3600, 1000),  # 1000 per an hour
    PerRequestThrottlingValidator(60, 20),  # 20 per an minute
)



import re
import json
from json import JSONDecodeError

from ipware.ip import get_ip
from jsonfield import JSONField

from django.conf import settings as django_settings
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.template.defaultfilters import truncatechars
from django.urls import resolve, reverse
from django.urls.exceptions import Resolver404
from django.utils import timezone
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import localtime

try:
    from django.core.urlresolvers import get_resolver
except ImportError:
    from django.urls import get_resolver

from security.config import settings
from security.utils import get_headers, remove_nul_from_string, regex_sub_groups_global, is_base_collection


try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except ImportError:
    from django.contrib.contenttypes.generic import GenericForeignKey


try:
    from pyston.filters.default_filters import CaseSensitiveStringFieldFilter
except ImportError:
    CaseSensitiveStringFieldFilter = object


def display_json(value, indent=4):
    return json.dumps(value, indent=indent, ensure_ascii=False, cls=DjangoJSONEncoder)


def hide_sensitive_data_body(content):
    for pattern in settings.HIDE_SENSITIVE_DATA_PATTERNS.get('BODY', ()):
        content = regex_sub_groups_global(pattern, settings.SENSITIVE_DATA_REPLACEMENT, content)
    return content


def hide_sensitive_data_headers(headers):
    headers = dict(headers)
    for pattern in settings.HIDE_SENSITIVE_DATA_PATTERNS.get('HEADERS', ()):
        for header_name, header in headers.items():
            if re.match(pattern, header_name, re.IGNORECASE):
                headers[header_name] = settings.SENSITIVE_DATA_REPLACEMENT
    return headers


def hide_sensitive_data_queries(queries):
    queries = dict(queries)
    for pattern in settings.HIDE_SENSITIVE_DATA_PATTERNS.get('QUERIES', ()):
        for query_name, query in queries.items():
            if re.match(pattern, query_name, re.IGNORECASE):
                queries[query_name] = (
                    len(query) * [settings.SENSITIVE_DATA_REPLACEMENT] if is_base_collection(query)
                    else settings.SENSITIVE_DATA_REPLACEMENT
                )
    return queries


def truncate_json_data(data):
    if isinstance(data, dict):
        return {key: truncate_json_data(val) for key, val in data.items()}
    elif isinstance(data, list):
        return [truncate_json_data(val) for val in data]
    elif isinstance(data, str):
        return truncatechars(data, settings.LOG_JSON_STRING_LENGTH)
    else:
        return data


def truncate_body(content, max_lenght):
    content = force_text(content, errors='replace')
    if len(content) > max_lenght:
        try:
            json_content = json.loads(content)
            return (
                json.dumps(truncate_json_data(json_content))
                if isinstance(json_content, (dict, list)) and settings.LOG_JSON_STRING_LENGTH is not None
                else content[:max_lenght + 1]
            )
        except JSONDecodeError:
            return content[:max_lenght + 1]
    else:
        return content


def clean_body(body, max_lenght):
    cleaned_body = truncatechars(
        truncate_body(body, max_lenght), max_lenght + len(settings.SENSITIVE_DATA_REPLACEMENT)
    ) if max_lenght is not None else str(body)
    cleaned_body = hide_sensitive_data_body(remove_nul_from_string(cleaned_body)) if cleaned_body else cleaned_body
    cleaned_body = truncatechars(cleaned_body, max_lenght) if max_lenght else cleaned_body
    return cleaned_body


def clean_headers(headers):
    return hide_sensitive_data_headers(headers) if headers else headers


def clean_queries(queries):
    return hide_sensitive_data_queries(queries) if queries else queries


class InputLoggedRequestManager(models.Manager):
    """
    Create new LoggedRequest instance from HTTP request
    """

    def prepare_from_request(self, request):
        user = hasattr(request, 'user') and request.user.is_authenticated and request.user or None
        path = truncatechars(request.path, 200)

        try:
            slug = resolve(request.path_info, getattr(request, 'urlconf', None)).view_name
        except Resolver404:
            slug = None

        return self.model(
            request_headers=clean_headers(get_headers(request)),
            request_body=clean_body(request.body, settings.LOG_REQUEST_BODY_LENGTH),
            user=user,
            method=request.method.upper()[:7],
            host=request.get_host(),
            path=path,
            queries=clean_queries(request.GET.dict()),
            is_secure=request.is_secure(),
            ip=get_ip(request),
            request_timestamp=timezone.now(),
            slug=slug
        )


class LoggedRequest(models.Model):

    INCOMPLETE = 0
    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50

    STATUS_CHOICES = (
        (INCOMPLETE, _('Incomplete')),
        (INFO, _('Info')),
        (WARNING, _('Warning')),
        (ERROR, _('Error')),
        (DEBUG, _('Debug')),
        (CRITICAL, _('Critical')),
    )

    host = models.CharField(_('host'), max_length=255, null=False, blank=False, db_index=True)
    host._filter = CaseSensitiveStringFieldFilter
    method = models.SlugField(_('method'), max_length=7, null=False, blank=False, db_index=True)
    path = models.CharField(_('URL path'), max_length=255, null=False, blank=False, db_index=True)
    path._filter = CaseSensitiveStringFieldFilter
    queries = JSONField(_('queries'), null=True, blank=True)
    is_secure = models.BooleanField(_('HTTPS connection'), default=False, null=False, blank=False)
    slug = models.SlugField(_('slug'), null=True, blank=True, db_index=True, max_length=255)

    # Request information
    request_timestamp = models.DateTimeField(_('request timestamp'), null=False, blank=False, db_index=True)
    request_headers = JSONField(_('request headers'), null=True, blank=True)
    request_body = models.TextField(_('request body'), null=False, blank=True)

    # Response information
    response_timestamp = models.DateTimeField(_('response timestamp'), null=True, blank=True)
    response_code = models.PositiveSmallIntegerField(_('response code'), null=True, blank=True)
    response_headers = JSONField(_('response headers'), null=True, blank=True)
    response_body = models.TextField(_('response body'), null=True, blank=True)

    status = models.PositiveSmallIntegerField(_('status'), choices=STATUS_CHOICES, null=False, blank=False,
                                              default=INCOMPLETE)
    error_description = models.TextField(_('error description'), null=True, blank=True)
    exception_name = models.CharField(_('exception name'), null=True, blank=True, max_length=255)

    @classmethod
    def get_status(cls, status_code):
        if status_code >= 500:
            return LoggedRequest.ERROR
        elif status_code >= 400:
            return LoggedRequest.WARNING
        else:
            return LoggedRequest.INFO

    def short_queries(self):
        return truncatechars(display_json(self.queries, indent=0), 50)
    short_queries.short_description = _('queries')
    short_queries.filter_by = 'queries'

    def short_request_headers(self):
        return truncatechars(display_json(self.request_headers, indent=0), 50)
    short_request_headers.short_description = _('request headers')
    short_request_headers.filter_by = 'request_headers'

    def response_time(self):
        return (self.response_timestamp - self.request_timestamp).total_seconds() if self.response_timestamp else None
    response_time.short_description = _('Response time')

    def short_path(self):
        return truncatechars(self.path, 50)
    short_path.short_description = _('Path')
    short_path.filter_by = 'path'
    short_path.order_by = 'path'

    def short_response_body(self):
        return truncatechars(self.response_body, 50)
    short_response_body.short_description = _('response body')
    short_response_body.filter_by = 'response_body'

    def short_request_body(self):
        return truncatechars(self.request_body, 50)
    short_request_body.short_description = _('request body')
    short_request_body.filter_by = 'request_body'

    def __str__(self):
        return ' '.join(
            (force_text(v) for v in (
                self.slug, self.response_code, localtime(self.request_timestamp.replace(microsecond=0))
            ) if v)
        )

    class Meta:
        abstract = True


class InputLoggedRequest(LoggedRequest):

    COMMON_REQUEST = 1
    THROTTLED_REQUEST = 2
    SUCCESSFUL_LOGIN_REQUEST = 3
    UNSUCCESSFUL_LOGIN_REQUEST = 4
    SUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST = 5
    UNSUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST = 6

    TYPE_CHOICES = (
        (COMMON_REQUEST, _('Common request')),
        (THROTTLED_REQUEST, _('Throttled request')),
        (SUCCESSFUL_LOGIN_REQUEST, _('Successful login request')),
        (UNSUCCESSFUL_LOGIN_REQUEST, _('Unsuccessful login request')),
        (SUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST, _('Successful two factor code verification request')),
        (UNSUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST, _('Unsuccessful two factor code verification request')),
    )

    user = models.ForeignKey(django_settings.AUTH_USER_MODEL, verbose_name=_('user'), null=True, blank=True,
                             on_delete=models.SET_NULL)
    ip = models.GenericIPAddressField(_('IP address'), null=False, blank=False)
    type = models.PositiveSmallIntegerField(_('type'), choices=TYPE_CHOICES, default=COMMON_REQUEST, null=False,
                                            blank=False)

    objects = InputLoggedRequestManager()

    def update_from_response(self, response):
        self.response_timestamp = timezone.now()
        self.status = self.get_status(response.status_code)
        self.response_code = response.status_code
        self.response_headers = clean_headers(dict(response.items()))

        if (not response.streaming and settings.LOG_RESPONSE_BODY_CONTENT_TYPES is not None and
                response.get('content-type', '').split(';')[0] in settings.LOG_RESPONSE_BODY_CONTENT_TYPES):
            response_body = clean_body(response.content, settings.LOG_RESPONSE_BODY_LENGTH)
        else:
            response_body = ''

        self.response_body = response_body

    class Meta:
        verbose_name = _('input logged request')
        verbose_name_plural = _('input logged requests')
        ordering = ('-request_timestamp',)


class OutputLoggedRequest(LoggedRequest):

    input_logged_request = models.ForeignKey(InputLoggedRequest, verbose_name=_('input logged request'), null=True,
                                             blank=True, on_delete=models.SET_NULL,
                                             related_name='output_logged_requests')

    class Meta:
        verbose_name = _('output logged request')
        verbose_name_plural = _('output logged requests')
        ordering = ('-request_timestamp',)


class OutputLoggedRequestRelatedObjects(models.Model):

    output_logged_request = models.ForeignKey(OutputLoggedRequest, verbose_name=_('output logged requests'), null=False,
                                              blank=False, related_name='related_objects', on_delete=models.CASCADE)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')

    def display_object(self, request):
        from is_core.utils import render_model_object_with_link

        return render_model_object_with_link(request, self.content_object) if self.content_object else None
    display_object.short_description = _('object')


class CommandLog(models.Model):
    """
    Represents a log of a command run.

    Attributes:
        start: Date and time when command was started.
        stop: Date and time when command finished.
        command_name: Name of the command.
        command_options: Arguments/options the command was run with.
        executed_from_command_line: Flag that indicates if command was run from the command line.
        output: Standard and error output of the command.
        is_successful: Flag that indicates if command finished successfully.
    """
    start = models.DateTimeField(blank=False, null=False, editable=False, verbose_name=_('start'))
    stop = models.DateTimeField(blank=True, null=True, editable=False, verbose_name=_('stop'))
    command_name = models.CharField(max_length=250, blank=False, null=False, editable=False, db_index=True,
                                    verbose_name=_('command name'))
    command_options = models.TextField(blank=False, null=False, editable=False, verbose_name=_('command options'))
    executed_from_command_line = models.BooleanField(blank=False, null=False, default=False, editable=False,
                                                     verbose_name=_('executed from command line'))
    output = models.TextField(blank=True, null=True, editable=False, verbose_name=_('command output'))
    is_successful = models.BooleanField(blank=False, null=False, default=False, editable=False,
                                        verbose_name=_('finished successfully'))

    class Meta:
        verbose_name = _('command log')
        verbose_name_plural = _('command logs')
        ordering = ('-start',)



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='loggedrequest',
            name='request_timestamp',
            field=models.DateTimeField(verbose_name='Request timestamp', db_index=True),
        ),
    ]



from django.core.management.base import CommandError

from security.models import InputLoggedRequest, OutputLoggedRequest
from security.utils import PurgeLogsBaseCommand


class Command(PurgeLogsBaseCommand):

    timestamp_field = 'request_timestamp'

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument('--type', action='store', dest='type', default='input',
                            help='Tells Django what type of requests should be removed (input/output).')

    def handle(self, expiration, type, **options):
        self.type = type
        super().handle(expiration, **options)

    def get_model(self):
        if self.type == 'input':
            return InputLoggedRequest
        elif self.type == 'output':
            return OutputLoggedRequest
        else:
            raise CommandError('Type can be only input or output')



import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import jsonfield.fields


def update_content_types(apps, schema_editor):
    ContentType = apps.get_model('contenttypes', 'ContentType')
    ContentType.objects.filter(app_label='security', model='loggedrequest').update(model='inputloggedrequest')


def migrate_request_status(apps, schema_editor):
    InputLoggedRequest = apps.get_model('security', 'InputLoggedRequest')
    for old_status, new_status in ((1, 20), (2, 30), (3, 50)):
        InputLoggedRequest.objects.filter(status=old_status).update(status=new_status)


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('security', '0002_auto_20151014_2159'),
    ]

    operations = [
        # Migrate LoggedRequest to InputLoggedRequest
        migrations.RenameModel('LoggedRequest', 'InputLoggedRequest'),
        migrations.RunPython(update_content_types),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='request_body',
            field=models.TextField(blank=True, verbose_name='request body'),
        ),
        migrations.RunPython(migrate_request_status),
        # Rename and alter InputLoggedRequest.headers field
        migrations.RenameField(
            model_name='inputloggedrequest',
            old_name='headers',
            new_name='request_headers',
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='request_headers',
            field=jsonfield.fields.JSONField(blank=True, default='null', help_text='Enter a valid JSON object',
                                             null=True, verbose_name='request headers'),
        ),
        # Alter other InputLoggedRequest fields
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    to=settings.AUTH_USER_MODEL, verbose_name='user'),
        ),
        migrations.AlterModelOptions(
            name='inputloggedrequest',
            options={'ordering': ('-request_timestamp',), 'verbose_name': 'input logged request',
                     'verbose_name_plural': 'input logged requests'},
        ),
        migrations.AddField(
            model_name='inputloggedrequest',
            name='exception_name',
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name='exception name'),
        ),
        migrations.AddField(
            model_name='inputloggedrequest',
            name='host',
            field=models.CharField(default='empty', max_length=255, verbose_name='host'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='inputloggedrequest',
            name='response_headers',
            field=jsonfield.fields.JSONField(blank=True, default='null', help_text='Enter a valid JSON object',
                                             null=True, verbose_name='response headers'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='error_description',
            field=models.TextField(blank=True, null=True, verbose_name='error description'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='method',
            field=models.CharField(max_length=7, verbose_name='method'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='queries',
            field=jsonfield.fields.JSONField(blank=True, default='null', help_text='Enter a valid JSON object',
                                             null=True, verbose_name='queries'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='request_body',
            field=models.TextField(blank=True, verbose_name='request body'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='request_timestamp',
            field=models.DateTimeField(db_index=True, verbose_name='request timestamp'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='response_body',
            field=models.TextField(blank=True, null=True, verbose_name='response body'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='response_code',
            field=models.PositiveSmallIntegerField(blank=True, null=True, verbose_name='response code'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='response_timestamp',
            field=models.DateTimeField(blank=True, null=True, verbose_name='response timestamp'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='status',
            field=models.PositiveSmallIntegerField(
                choices=[(20, 'Info'), (30, 'Warning'), (40, 'Error'), (10, 'Debug'), (50, 'Critical')],
                verbose_name='status'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='type',
            field=models.PositiveSmallIntegerField(
                choices=[(1, 'Common request'), (2, 'Throttled request'), (3, 'Successful login request'),
                         (4, 'Unsuccessful login request')], default=1, verbose_name='type'),
        ),

        migrations.CreateModel(
            name='OutputLoggedRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('host', models.CharField(max_length=255, verbose_name='host')),
                ('method', models.CharField(max_length=7, verbose_name='method')),
                ('path', models.CharField(max_length=255, verbose_name='URL path')),
                ('queries',
                 jsonfield.fields.JSONField(blank=True, default='null', help_text='Enter a valid JSON object',
                                            null=True, verbose_name='queries')),
                ('is_secure', models.BooleanField(default=False, verbose_name='HTTPS connection')),
                ('request_timestamp', models.DateTimeField(db_index=True, verbose_name='request timestamp')),
                ('request_headers',
                 jsonfield.fields.JSONField(blank=True, default='null', help_text='Enter a valid JSON object',
                                            null=True, verbose_name='request headers')),
                ('request_body', models.TextField(blank=True, verbose_name='request body')),
                ('response_timestamp', models.DateTimeField(blank=True, null=True, verbose_name='response timestamp')),
                (
                'response_code', models.PositiveSmallIntegerField(blank=True, null=True, verbose_name='response code')),
                ('response_headers',
                 jsonfield.fields.JSONField(blank=True, default='null', help_text='Enter a valid JSON object',
                                            null=True, verbose_name='response headers')),
                ('response_body', models.TextField(blank=True, null=True, verbose_name='response body')),
                ('status', models.PositiveSmallIntegerField(
                    choices=[(20, 'Info'), (30, 'Warning'), (40, 'Error'), (10, 'Debug'), (50, 'Critical')],
                    verbose_name='status')),
                ('error_description', models.TextField(blank=True, null=True, verbose_name='error description')),
                ('exception_name',
                 models.CharField(blank=True, max_length=255, null=True, verbose_name='exception name')),
                ('slug', models.SlugField(blank=True, null=True, verbose_name='slug')),
            ],
            options={
                'verbose_name': 'output logged request',
                'verbose_name_plural': 'output logged requests',
                'ordering': ('-request_timestamp',),
            },
        ),
        migrations.CreateModel(
            name='OutputLoggedRequestRelatedObjects',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField()),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('output_logged_request',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='related_objects',
                                   to='security.OutputLoggedRequest', verbose_name='output logged requests')),
            ],
        )
    ]



from django.utils.translation import ugettext_lazy as _

from is_core.generic_views.inlines.inline_form_views import TabularInlineFormView

from security.cores import InputRequestsLogISCore as OriginInputRequestsLogISCore
from security.reversion_log.models import InputRequestRevision


class RequestRevisionTabularInlineFormView(TabularInlineFormView):

    model = InputRequestRevision


class InputRequestsLogISCore(OriginInputRequestsLogISCore):

    form_fieldsets = (
        (_('Request'), {'fields': ('request_timestamp', 'method', 'path', 'queries', 'request_headers_code',
                                   'request_body_code', 'is_secure')}),
        (_('Response'), {'fields': ('response_timestamp', 'response_code', 'status', 'response_headers_code',
                                    'response_body_code', 'type', 'error_description_code')}),
        (_('User information'), {'fields': ('user', 'ip')}),
        (_('Extra information'), {'fields': ('response_time', 'output_logged_requests',)}),
        (_('Revisions'), {'inline_view': RequestRevisionTabularInlineFormView}),
    )
    abstract = True



import io
import json

from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase, override_settings
from django.db import transaction

import responses

from security.models import InputLoggedRequest, OutputLoggedRequest, CommandLog
from security.management import call_command
from security.transport import security_requests as requests
from security.transport.transaction import atomic_log

from germanium.annotations import data_provider
from germanium.test_cases.client import ClientTestCase
from germanium.tools import (
    assert_in, assert_not_in, assert_equal, assert_true, assert_false, assert_http_redirect, assert_http_bad_request,
    assert_http_application_error, assert_http_ok, assert_http_not_found, assert_raises, assert_http_too_many_requests
)


class TestException(Exception):
    pass


class BaseTestCaseMixin:

    def create_user(self):
        return User.objects._create_user('test', 'test@test.cz', 'test', is_staff=True, is_superuser=True)


class SecurityTestCase(BaseTestCaseMixin, ClientTestCase):

    def test_every_request_should_be_logged(self):
        assert_equal(InputLoggedRequest.objects.count(), 0)
        self.get('/')
        assert_equal(InputLoggedRequest.objects.count(), 1)

    @data_provider('create_user')
    def test_data_change_should_be_connected_with_logged_request(self, user):
        assert_equal(InputLoggedRequest.objects.count(), 0)
        assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
        assert_equal(InputLoggedRequest.objects.count(), 1)
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(
            input_logged_request.input_logged_request_revisions.get().revision.version_set.get().content_type,
            ContentType.objects.get_for_model(User)
        )

    @data_provider('create_user')
    def test_input_logged_request_should_have_right_status(self, user):
        assert_http_ok(self.post('/admin/login/', data={'username': 'invalid', 'password': 'invalid'}))
        assert_equal(InputLoggedRequest.objects.first().status, InputLoggedRequest.INFO)
        assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
        assert_equal(InputLoggedRequest.objects.first().status, InputLoggedRequest.INFO)
        assert_raises(Exception, self.get, '/proxy/')
        assert_equal(InputLoggedRequest.objects.first().status, InputLoggedRequest.ERROR)
        assert_http_not_found(self.get('/404/'))
        assert_equal(InputLoggedRequest.objects.first().status, InputLoggedRequest.WARNING)

    @override_settings(SECURITY_LOG_REQUEST_IGNORE_IP=('127.0.0.1',))
    def test_ignored_client_ip_should_not_be_logged(self):
        assert_equal(InputLoggedRequest.objects.count(), 0)
        self.get('/')
        assert_equal(InputLoggedRequest.objects.count(), 0)

    @override_settings(SECURITY_LOG_REQUEST_IGNORE_URL_PATHS=('/',))
    def test_ignored_request_path_should_not_be_logged(self):
        assert_equal(InputLoggedRequest.objects.count(), 0)
        self.get('/')
        assert_equal(InputLoggedRequest.objects.count(), 0)

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=10)
    def test_request_body_should_be_truncated(self):
        self.post('/admin/login/', data={'username': 20 * 'a', 'password': 20 * 'b'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(len(input_logged_request.request_body), 10)
        assert_true(input_logged_request.request_body.endswith('...'))

    @override_settings(SECURITY_LOG_RESPONSE_BODY_LENGTH=10)
    def test_response_body_should_be_truncated(self):
        self.post('/admin/login/', data={'username': 20 * 'a', 'password': 20 * 'b'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(len(input_logged_request.response_body), 10)
        assert_true(input_logged_request.response_body.endswith('...'))

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=None)
    def test_request_body_truncation_should_be_turned_off(self):
        self.post('/admin/login/', data={'username': 2000 * 'a', 'password': 2000 * 'b'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(len(input_logged_request.request_body), 4183)
        assert_false(input_logged_request.request_body.endswith('...'))

    @override_settings(SECURITY_LOG_RESPONSE_BODY_LENGTH=None)
    def test_response_body_truncation_should_be_turned_off(self):
        resp = self.post('/admin/login/', data={'username': 20 * 'a', 'password': 20 * 'b'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(input_logged_request.response_body, str(resp.content))
        assert_false(input_logged_request.response_body.endswith('...'))

    @override_settings(SECURITY_LOG_RESPONSE_BODY_CONTENT_TYPES=())
    def test_not_allowed_content_type_should_not_be_logged(self):
        self.get('/')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_false(input_logged_request.response_body)

    @override_settings(SECURITY_LOG_RESPONSE_BODY_CONTENT_TYPES=None)
    def test_allowed_content_type_should_be_turned_off(self):
        self.get('/')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_false(input_logged_request.response_body)

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=100, SECURITY_LOG_JSON_STRING_LENGTH=10)
    def test_json_request_should_be_truncated_with_another_method(self):
        self.c.post('/admin/login/', data=json.dumps({'a': 50 * 'a', 'b': 50 * 'b'}),
                    content_type='application/json')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(
            json.loads(input_logged_request.request_body),
            json.loads('{"a": "aaaaaaa...", "b": "bbbbbbb..."}')
        )
        assert_false(input_logged_request.request_body.endswith('...'))

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=50, SECURITY_LOG_JSON_STRING_LENGTH=None)
    def test_json_request_should_not_be_truncated_with_another_method(self):
        self.c.post('/admin/login/', data=json.dumps({'a': 50 * 'a'}),
                    content_type='application/json')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(input_logged_request.request_body, '{"a": "' + 40* 'a' + '...')
        assert_true(input_logged_request.request_body.endswith('...'))

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=100, SECURITY_LOG_JSON_STRING_LENGTH=10)
    def test_json_request_should_be_truncated_with_another_method_and_standard_method_too(self):
        self.c.post('/admin/login/', data=json.dumps({50 * 'a': 50 * 'a', 50 * 'b': 50 * 'b'}),
                    content_type='application/json')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(len(input_logged_request.request_body), 100)
        assert_true(input_logged_request.request_body.endswith('...'))

    def test_response_with_exception_should_be_logged(self):
        assert_equal(InputLoggedRequest.objects.count(), 0)
        assert_raises(Exception, self.get, '/proxy/')
        assert_equal(InputLoggedRequest.objects.count(), 1)

    @responses.activate
    def test_response_with_exception_should_be_logged(self):
        responses.add(responses.GET, 'http://test.cz', body='test')
        assert_equal(self.get('/proxy/?url=http://test.cz').content, b'test')
        assert_equal(InputLoggedRequest.objects.count(), 1)
        assert_equal(OutputLoggedRequest.objects.count(), 1)

    @responses.activate
    @data_provider('create_user')
    def test_output_logged_request_should_be_related_with_object(self, user):
        assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
        responses.add(responses.GET, 'http://test.cz', body='test')
        assert_equal(self.get('/proxy/?url=http://test.cz').content, b'test')
        assert_equal(InputLoggedRequest.objects.count(), 2)
        assert_equal(OutputLoggedRequest.objects.count(), 1)
        output_logged_request = OutputLoggedRequest.objects.get()
        assert_equal(output_logged_request.related_objects.get().content_object, user)

    def test_sensitive_data_body_in_json_should_be_hidden(self):
        self.c.post('/admin/login/', data=json.dumps({'username': 'test', 'password': 'secret-password'}),
                    content_type='application/json')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_in('"password": "[Filtered]"', input_logged_request.request_body)
        assert_not_in('"password": "secret-password"', input_logged_request.request_body)

    def test_sensitive_data_body_in_raw_form_should_be_hidden(self):
        self.post('/admin/login/', data={'username': 'test', 'password': 'secret-password\nddd'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_in('[Filtered]', input_logged_request.request_body)

    @data_provider('create_user')
    def test_sensitive_headers_should_be_hidden(self, user):
        assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(input_logged_request.request_headers['COOKIE'], '[Filtered]')

    def test_sensitive_queries_should_be_hidden(self):
        self.get('/?token=test')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(input_logged_request.queries['token'], '[Filtered]')

    @data_provider('create_user')
    @override_settings(SECURITY_SENSITIVE_DATA_REPLACEMENT='(Filtered)')
    def test_sensitive_replacement_should_be_changed(self, user):
        assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(input_logged_request.request_headers['COOKIE'], '(Filtered)')

    @override_settings(SECURITY_COMMAND_LOG_EXCLUDED_COMMANDS=())
    def test_command_should_be_logged(self):
        assert_equal(CommandLog.objects.count(), 0)
        call_command('showmigrations', stdout=io.StringIO(), stderr=io.StringIO())
        assert_equal(CommandLog.objects.count(), 1)

    @override_settings(SECURITY_COMMAND_LOG_EXCLUDED_COMMANDS=('showmigrations',))
    def test_excluded_command_should_not_be_logged(self):
        assert_equal(CommandLog.objects.count(), 0)
        call_command('showmigrations', stdout=io.StringIO(), stderr=io.StringIO())
        assert_equal(CommandLog.objects.count(), 0)

    def test_throttling_should_be_raised(self):
        for _ in range(20):
            assert_http_redirect(self.get('/admin/'))
        assert_http_too_many_requests(self.get('/admin/'))

    @override_settings(SECURITY_DEFAULT_THROTTLING_VALIDATORS_PATH='apps.test_security.tests.throttling_validators')
    def test_throttling_configuration_should_be_changed_via_settings(self):
        for _ in range(2):
            assert_http_redirect(self.get('/admin/'))
        assert_http_too_many_requests(self.get('/admin/'))

    def test_decorated_view_with_hide_request_body_should_not_log_request_body(self):
        self.post('/hide-request-body/', data={'a': 20 * 'a', 'b': 20 * 'b'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_false(input_logged_request.request_body)

    def test_decorated_view_with_log_exempt_should_not_log_request(self):
        self.get('/log-exempt/')
        assert_equal(InputLoggedRequest.objects.count(), 0)

    def test_decorated_view_with_throttling_exempt_should_not_raise_throttling_exception(self):
        for _ in range(20):
            assert_http_ok(self.get('/throttling-exempt/'))
        assert_http_ok(self.get('/throttling-exempt/'))

    def test_decorated_view_with_throttling_should_raise_throttling_exception(self):
        assert_http_ok(self.get('/extra-throttling/'))
        assert_http_too_many_requests(self.get('/extra-throttling/'))

    @responses.activate
    def test_output_logged_requests_with_atomic_block_should_not_be_logged_if_exception_is_raised(self):
        responses.add(responses.GET, 'http://test.cz', body='test')
        with assert_raises(TestException):
            with transaction.atomic():
                requests.get('http://test.cz')
                assert_equal(OutputLoggedRequest.objects.count(), 1)
                raise TestException
        assert_equal(OutputLoggedRequest.objects.count(), 0)

    @responses.activate
    def test_output_logged_requests_with_atomic_and_log_atomic_block_should_be_logged_if_exception_is_raised(self):
        responses.add(responses.GET, 'http://test.cz', body='test')
        with assert_raises(TestException):
            with atomic_log():
                with transaction.atomic():
                    requests.get('http://test.cz')
                    assert_equal(OutputLoggedRequest.objects.count(), 0)
                    raise TestException
        assert_equal(OutputLoggedRequest.objects.count(), 1)

    @responses.activate
    def test_output_logged_requests_with_atomic_and_log_atomic_block_should_be_nested(self):
        responses.add(responses.GET, 'http://test.cz', body='test')
        with assert_raises(TestException):
            with atomic_log():
                with transaction.atomic():
                    requests.get('http://test.cz')
                    with atomic_log():
                        with transaction.atomic():
                            requests.get('http://test.cz')
                    assert_equal(OutputLoggedRequest.objects.count(), 0)
                    raise TestException
        assert_equal(OutputLoggedRequest.objects.count(), 2)

    @responses.activate
    def test_response_sensitive_data_body_in_json_should_be_hidden(self):
        responses.add(responses.POST, 'http://test.cz', body='test')
        requests.post('http://test.cz', data=json.dumps({'password': 'secret-password'}))
        output_logged_requst = OutputLoggedRequest.objects.get()
        assert_in('"password": "[Filtered]"', output_logged_requst.request_body)
        assert_not_in('"password": "secret-password"', output_logged_requst.request_body)
        assert_in('"password": "secret-password"', responses.calls[0].request.body)
        assert_not_in('"password": "[Filtered]"', responses.calls[0].request.body)

    @responses.activate
    def test_response_sensitive_headers_should_be_hidden(self):
        responses.add(responses.POST, 'http://test.cz', body='test')
        requests.post('http://test.cz', headers={'token': 'secret'})
        output_logged_request = OutputLoggedRequest.objects.get()
        assert_equal(output_logged_request.request_headers['token'], '[Filtered]')
        assert_equal(responses.calls[0].request.headers['token'], 'secret')

    @responses.activate
    def test_response_sensitive_params_data_should_be_hidden(self):
        responses.add(responses.POST, 'http://test.cz', body='test')
        requests.post('http://test.cz', params={'token': 'secret'})
        output_logged_request = OutputLoggedRequest.objects.get()
        assert_equal(output_logged_request.queries['token'], '[Filtered]')
        assert_equal(responses.calls[0].request.url, 'http://test.cz/?token=secret')

    @responses.activate
    def test_response_more_sensitive_params_data_should_be_hidden(self):
        responses.add(responses.POST, 'http://test.cz', body='test')
        requests.post('http://test.cz', params={'token': ['secret', 'secret2']})
        output_logged_request = OutputLoggedRequest.objects.get()
        assert_equal(output_logged_request.queries['token'], ['[Filtered]', '[Filtered]'])
        assert_equal(responses.calls[0].request.url, 'http://test.cz/?token=secret&token=secret2')

    @responses.activate
    def test_response_sensitive_params_and_url_query_together_data_should_be_logged(self):
        responses.add(responses.POST, 'http://test.cz', body='test')
        requests.post('http://test.cz?a=1&a=2', params={'b': '6', 'a': '3', 'c': ['5']})
        output_logged_request = OutputLoggedRequest.objects.get()
        assert_equal(output_logged_request.queries, {'b': '6', 'a': ['3', '1', '2'], 'c': '5'})



import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import jsonfield.fields


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='LoggedRequest',
            fields=[
                ('id', models.AutoField(primary_key=True, auto_created=True, serialize=False, verbose_name='ID')),
                ('request_timestamp', models.DateTimeField(verbose_name='Request timestamp')),
                ('method', models.CharField(verbose_name='Method', max_length=7)),
                ('path', models.CharField(verbose_name='URL path', max_length=255)),
                ('queries',
                 jsonfield.fields.JSONField(help_text='Enter a valid JSON object', verbose_name='Queries', null=True,
                                            default='null', blank=True)),
                ('headers',
                 jsonfield.fields.JSONField(help_text='Enter a valid JSON object', verbose_name='Headers', null=True,
                                            default='null', blank=True)),
                ('request_body', models.TextField(verbose_name='Request body', blank=True)),
                ('is_secure', models.BooleanField(verbose_name='HTTPS connection', default=False)),
                ('response_timestamp', models.DateTimeField(verbose_name='Response timestamp')),
                ('response_code', models.PositiveSmallIntegerField(verbose_name='Response code')),
                ('status', models.PositiveSmallIntegerField(verbose_name='Status',
                                                            choices=[(1, 'Fine'), (2, 'Warning'), (3, 'Error')])),
                ('type', models.PositiveSmallIntegerField(verbose_name='Request type', default=1,
                                                          choices=[(1, 'Common request'), (2, 'Throttled request'),
                                                                   (3, 'Successful login request'),
                                                                   (4, 'Unsuccessful login request')])),
                ('response_body', models.TextField(verbose_name='Response body', blank=True)),
                ('error_description', models.TextField(verbose_name='Error description', null=True, blank=True)),
                ('ip', models.GenericIPAddressField(verbose_name='IP address')),
                ('user',
                 models.ForeignKey(null=True, to=settings.AUTH_USER_MODEL, on_delete=django.db.models.deletion.SET_NULL,
                                   blank=True)),
            ],
            options={
                'verbose_name': 'Logged request',
                'ordering': ('-request_timestamp',),
                'verbose_name_plural': 'Logged requests',
            },
        ),
    ]



from django.db.transaction import get_connection

from reversion import revisions as reversion

from security.middleware import MiddlewareMixin
from security.reversion_log.models import InputRequestRevision


class RevisionLogMiddleware(MiddlewareMixin):

    def process_request(self, request):
        def create_revision_request_log(revision):
            if getattr(request, 'input_logged_request', False):
                InputRequestRevision.objects.create(logged_request=request.input_logged_request, revision=revision)

        reversion.add_callback(create_revision_request_log)



import jsonfield.fields

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0003_auto_20160509_1622'),
    ]

    operations = [
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='queries',
            field=jsonfield.fields.JSONField(blank=True, null=True, verbose_name='queries'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='request_headers',
            field=jsonfield.fields.JSONField(blank=True, null=True, verbose_name='request headers'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='response_headers',
            field=jsonfield.fields.JSONField(blank=True, null=True, verbose_name='response headers'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='queries',
            field=jsonfield.fields.JSONField(blank=True, null=True, verbose_name='queries'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='request_headers',
            field=jsonfield.fields.JSONField(blank=True, null=True, verbose_name='request headers'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='response_headers',
            field=jsonfield.fields.JSONField(blank=True, null=True, verbose_name='response headers'),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reversion', '__first__'),
        ('security', '0003_auto_20160509_1622'),
    ]

    operations = [
        migrations.CreateModel(
            name='InputRequestRevision',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('logged_request', models.ForeignKey(verbose_name='logged request', to='security.InputLoggedRequest',
                                                     on_delete=models.CASCADE)),
                ('revision', models.OneToOneField(verbose_name='revision', to='reversion.Revision',
                                                  on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'Logged request revision',
                'verbose_name_plural': 'Logged requests revisions',
            },
        ),
    ]



import traceback

from ipware.ip import get_ip

from django import http
from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured
from django.db.transaction import get_connection
from django.urls import is_valid_path, get_callable, resolve
from django.utils.encoding import force_text

from .config import settings
from .exception import ThrottlingException
from .models import InputLoggedRequest
from .utils import get_throttling_validators, get_view_from_request_or_none

try:
    from importlib import import_module
except ImportError:  # For Django < 1.8
    from django.utils.importlib import import_module


class MiddlewareMixin:

    def __init__(self, get_response=None):
        self.get_response = get_response
        super().__init__()

    def __call__(self, request):
        response = None
        if hasattr(self, 'process_request'):
            response = self.process_request(request)
        response = response or self.get_response(request)
        if hasattr(self, 'process_response'):
            response = self.process_response(request, response)
        return response


if 'security.reversion_log' in django_settings.INSTALLED_APPS:
    if 'reversion' not in django_settings.INSTALLED_APPS:
        raise ImproperlyConfigured('For reversion log is necessary install "django-reversion"')

    # Supports two version of reversion library
    try:
        from reversion.signals import post_revision_commit

        def create_revision_request_log(sender, revision, versions, **kwargs):
            from security.reversion_log.models import InputRequestRevision

            connection = get_connection()
            if getattr(connection, 'input_logged_request', False):
                InputRequestRevision.objects.create(logged_request=connection.input_logged_request, revision=revision)

        post_revision_commit.connect(create_revision_request_log)
    except ImportError:
        pass


class LogMiddleware(MiddlewareMixin):

    response_redirect_class = http.HttpResponsePermanentRedirect

    def process_request(self, request):
        connection = get_connection()

        view = get_view_from_request_or_none(request)
        if (get_ip(request) not in settings.LOG_REQUEST_IGNORE_IP 
               and request.path not in settings.LOG_REQUEST_IGNORE_URL_PATHS
               and not getattr(view, 'log_exempt', False)):
            input_logged_request = InputLoggedRequest.objects.prepare_from_request(request)
            if getattr(view, 'hide_request_body', False):
                input_logged_request.request_body = ''
            input_logged_request.save()
            request.input_logged_request = input_logged_request
            connection.input_logged_request = input_logged_request

        output_logged_requests = getattr(connection, 'output_logged_requests', [])
        output_logged_requests.append([])
        request.output_logged_requests = output_logged_requests
        connection.output_logged_requests = output_logged_requests

        # Return a redirect if necessary
        if self.should_redirect_with_slash(request):
            return self.response_redirect_class(self.get_full_path_with_slash(request))

    def get_full_path_with_slash(self, request):
        """
        Return the full path of the request with a trailing slash appended.

        Raise a RuntimeError if settings.DEBUG is True and request.method is
        POST, PUT, or PATCH.
        """
        new_path = request.get_full_path(force_append_slash=True)
        if django_settings.DEBUG and request.method in {'POST', 'PUT', 'PATCH'}:
            raise RuntimeError(
                "You called this URL via %(method)s, but the URL doesn't end "
                "in a slash and you have SECURITY_APPEND_SLASH set. Django can't "
                "redirect to the slash URL while maintaining %(method)s data. "
                "Change your form to point to %(url)s (note the trailing "
                "slash), or set SECURITY_APPEND_SLASH=False in your Django settings." % {
                    'method': request.method,
                    'url': request.get_host() + new_path,
                }
            )
        return new_path

    def should_redirect_with_slash(self, request):
        """
        Return True if settings.APPEND_SLASH is True and appending a slash to
        the request path turns an invalid path into a valid one.
        """
        if settings.APPEND_SLASH and not request.get_full_path().endswith('/'):
            urlconf = getattr(request, 'urlconf', None)
            return (
                not is_valid_path(request.path_info, urlconf) and
                is_valid_path('{}/'.format(request.path_info), urlconf)
            )
        return False

    def _render_throttling(self, request, exception):
        return get_callable(settings.THROTTLING_FAILURE_VIEW)(request, exception)

    def process_view(self, request, callback, callback_args, callback_kwargs):
        if getattr(request, 'input_logged_request', False):

            throttling_validators = getattr(callback, 'throttling_validators', None)
            if throttling_validators is None:
                throttling_validators = get_throttling_validators('default_validators')

            try:
                for validator in throttling_validators:
                    validator.validate(request)
            except ThrottlingException as exception:
                return self.process_exception(request, exception)

    def process_response(self, request, response):
        input_logged_request = getattr(request, 'input_logged_request', None)
        if input_logged_request:
            input_logged_request.update_from_response(response)
            input_logged_request.save()

        output_logged_requests = (
            request.output_logged_requests.pop() if hasattr(request, 'output_logged_requests') else ()
        )
        [logged_request.create(input_logged_request) for logged_request in output_logged_requests]
        return response

    def process_exception(self, request, exception):
        if hasattr(request, 'input_logged_request'):
            logged_request = request.input_logged_request
            logged_request.error_description = traceback.format_exc()
            logged_request.exception_name = exception.__class__.__name__
            if isinstance(exception, ThrottlingException):
                logged_request.type = InputLoggedRequest.THROTTLED_REQUEST
                return self._render_throttling(request, exception)



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0005_auto_20161218_2224'),
    ]

    operations = [
        migrations.AddField(
            model_name='inputloggedrequest',
            name='slug',
            field=models.SlugField(blank=True, max_length=255, null=True, verbose_name='slug'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='slug',
            field=models.SlugField(blank=True, max_length=255, null=True, verbose_name='slug'),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0004_auto_20160530_1608'),
    ]

    operations = [
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='host',
            field=models.CharField(db_index=True, max_length=255, verbose_name='host'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='method',
            field=models.SlugField(max_length=7, verbose_name='method'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='path',
            field=models.CharField(db_index=True, max_length=255, verbose_name='URL path'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='host',
            field=models.CharField(db_index=True, max_length=255, verbose_name='host'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='method',
            field=models.SlugField(max_length=7, verbose_name='method'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='path',
            field=models.CharField(db_index=True, max_length=255, verbose_name='URL path'),
        ),
    ]



from django.shortcuts import render
from django.utils.encoding import force_text


def throttling_failure_view(request, exception):
    response = render(request, '429.html', {'description': force_text(exception)})
    response.status_code = 429
    return response



from django.db import DEFAULT_DB_ALIAS
from django.db.transaction import get_connection
from django.utils.decorators import ContextDecorator

from security.models import OutputLoggedRequest


class OutputLoggedRequestContext:
    """
    Data structure that stores data for creation OutputLoggedRequest model object
    """

    def __init__(self, data, related_objects=None):
        self.data = data
        self.related_objects = related_objects

    def create(self, input_logged_request=None):
        output_logged_request = OutputLoggedRequest.objects.create(
            input_logged_request=input_logged_request, **self.data
        )
        for obj in self.related_objects or ():
            if obj.__class__.objects.filter(pk=obj.pk).exists():
                output_logged_request.related_objects.create(content_object=obj)


class AtomicLog(ContextDecorator):
    """
    Context decorator that stores logged requests to database connections and inside exit method
    stores it to the database
    """

    def __init__(self, using):
        self.using = using

    def __enter__(self):
        connection = get_connection(self.using)
        output_logged_requests = getattr(connection, 'output_logged_requests', [])
        output_logged_requests.append([])
        connection.output_logged_requests = output_logged_requests

    def __exit__(self, exc_type, exc_value, traceback):
        connection = get_connection(self.using)
        output_logged_requests = connection.output_logged_requests.pop()
        if connection.output_logged_requests:
            connection.output_logged_requests[-1] += output_logged_requests
        else:
            [output_logged_request.create() for output_logged_request in output_logged_requests]


def atomic_log(using=None):
    """
    Decorator that surrounds atomic block, ensures that logged output requests will be stored inside database in case
    of DB rollback
    """
    if callable(using):
        return AtomicLog(DEFAULT_DB_ALIAS)(using)
    else:
        return AtomicLog(using)


def log_output_request(data, related_objects=None, using=None):
    """
    Helper for logging output requests
    :param data: dict of input attributes of OutputLoggedRequest model
    :param related_objects: objects that will be related to OutputLoggedRequest object
    :param using: database alias
    """
    if is_active_logged_requests(using):
        output_logged_requests = get_connection(using).output_logged_requests[-1]
        output_logged_requests.append(OutputLoggedRequestContext(data, related_objects))
    else:
        output_logged_request = OutputLoggedRequest.objects.create(**data)
        if related_objects:
            [output_logged_request.related_objects.create(content_object=obj) for obj in related_objects]


def is_active_logged_requests(using=None):
    """
    :param using: database alias
    :return: True if block of code is surrounded with atomic_log operator
    """
    return hasattr(get_connection(using), 'output_logged_requests') and get_connection(using).output_logged_requests



# Generated by Django 2.0 on 2018-04-18 09:12

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('security', '0008_auto_20180306_2254'),
    ]

    operations = [
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='status',
            field=models.PositiveSmallIntegerField(
                choices=[(0, 'Incomplete'), (20, 'Info'), (30, 'Warning'), (40, 'Error'), (10, 'Debug'),
                         (50, 'Critical')], default=0, verbose_name='status'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='input_logged_request',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='output_logged_requests', to='security.InputLoggedRequest',
                                    verbose_name='input logged request'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='status',
            field=models.PositiveSmallIntegerField(
                choices=[(0, 'Incomplete'), (20, 'Info'), (30, 'Warning'), (40, 'Error'), (10, 'Debug'),
                         (50, 'Critical')], default=0, verbose_name='status'),
        ),
    ]



from django.db import models
from django.utils.translation import ugettext_lazy as _

from reversion.models import Revision

from security.models import InputLoggedRequest


class InputRequestRevision(models.Model):

    logged_request = models.ForeignKey(InputLoggedRequest, verbose_name=_('logged request'), null=False, blank=False,
                                       on_delete=models.CASCADE, related_name='input_logged_request_revisions')
    revision = models.OneToOneField(Revision, verbose_name=_('revision'), null=False, blank=False,
                                    on_delete=models.CASCADE, related_name='input_logged_request_revision')

    def __str__(self):
        return ' #%s' % self.pk

    class Meta:
        verbose_name = _('Logged request revision')
        verbose_name_plural = _('Logged requests revisions')



import sys

from argparse import ArgumentParser

import django
from django.core.management import call_command as call_command_original
from django.core.management import execute_from_command_line as execute_from_command_line_original
from django.core.management import handle_default_options


def execute_from_command_line(argv=None):
    # some arguments must be processed before django setup
    parser = ArgumentParser()
    parser.add_argument('--settings')
    parser.add_argument('--pythonpath')
    options, args = parser.parse_known_args(argv[2:])
    handle_default_options(options)

    django.setup()

    from security.utils import CommandLogger
    return CommandLogger(
        command_function=lambda: execute_from_command_line_original(argv),
        command_name=argv[1],
        command_options=' '.join(argv[2:]),
        executed_from_command_line=True
    ).run()


def call_command(command_name, stdout=None, stderr=None, *args, **options):
    stdout = sys.stdout if stdout is None else stdout
    stderr = sys.stderr if stderr is None else stderr

    from security.utils import CommandLogger
    return CommandLogger(
        command_function=lambda: call_command_original(command_name, stdout=stdout, stderr=stderr, *args, **options),
        stdout=stdout,
        stderr=stderr,
        command_name=command_name,
        command_options=', '.join(
            list(args)+['{}={}'.format(k, v) for k, v in options.items()]
        )
    ).run()



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('security', '0007_outputloggedrequest_input_logged_request'),
    ]

    operations = [
        migrations.CreateModel(
            name='CommandLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start', models.DateTimeField(editable=False, verbose_name='start')),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='stop')),
                ('command_name',
                 models.CharField(db_index=True, editable=False, max_length=250, verbose_name='command name')),
                ('command_options', models.TextField(editable=False, verbose_name='command options')),
                ('executed_from_command_line',
                 models.BooleanField(default=False, editable=False, verbose_name='executed from command line')),
                ('output', models.TextField(blank=True, editable=False, null=True, verbose_name='command output')),
                ('is_successful',
                 models.BooleanField(default=False, editable=False, verbose_name='finished successfully')),
            ],
            options={
                'verbose_name': 'command log',
                'ordering': ('-start',),
                'verbose_name_plural': 'command logs',
            },
        ),
    ]



from django.contrib import admin
from django.urls import path

from apps.test_security.views import (
    proxy_view, hide_request_body_view, log_exempt_view, throttling_exempt_view, extra_throttling_view
)

admin.autodiscover()


urlpatterns = [
    path('admin/', admin.site.urls),
    path('proxy/', proxy_view),
    path('hide-request-body/', hide_request_body_view),
    path('log-exempt/', log_exempt_view),
    path('throttling-exempt/', throttling_exempt_view),
    path('extra-throttling/', extra_throttling_view),
]



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('security', '0006_auto_20171002_1415'),
    ]

    operations = [
        migrations.AddField(
            model_name='outputloggedrequest',
            name='input_logged_request',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    to='security.InputLoggedRequest', verbose_name='input logged request'),
        ),
    ]



# Generated by Django 2.0 on 2018-04-18 09:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('reversion_log', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='inputrequestrevision',
            name='logged_request',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='input_logged_request',
                                    to='security.InputLoggedRequest', verbose_name='logged request'),
        ),
        migrations.AlterField(
            model_name='inputrequestrevision',
            name='revision',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='input_logged_request',
                                       to='reversion.Revision', verbose_name='revision'),
        ),
    ]



from requests import *
from urllib.parse import parse_qs, urlparse

from django.utils import timezone
from django.utils.encoding import force_text

from security.config import settings
from security.models import LoggedRequest, clean_body, clean_headers, clean_queries
from security.utils import is_base_collection

from .transaction import log_output_request


def stringify_dict(d):
    def stringify(value):
        if isinstance(value, bytes):
            return force_text(value)
        elif isinstance(value, dict):
            return stringify_dict(value)
        else:
            return value

    return {k: stringify(v) for k, v in d.items()}


def prepare_body(body):
    return force_text(body, errors='replace')


def flat_params(params):
    return {
        k: v[0] if is_base_collection(v) and len(v) == 1 else v
        for k, v in params.items()
    }


def list_params(params):
    return {
        k: list(v) if is_base_collection(v) else [v]
        for k, v in params.items()
    }


def get_logged_params(url, params):
    params = list_params(params or {})
    parsed_params = parse_qs(urlparse(url).query)

    for k, v in parsed_params.items():
        if k not in params:
            params[k] = v
        else:
            params[k] += v

    return flat_params(params)


class SecuritySession(Session):

    def __init__(self, slug=None, related_objects=None):
        super(SecuritySession, self).__init__()
        self.slug = slug
        self.related_objects = related_objects

    def request(self, method, url, params=None, data=None, headers=None, cookies=None, files=None, auth=None,
                timeout=None, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None,
                json=None, slug=None, related_objects=None):

        parsed_url = urlparse(url)
        logged_kwargs = {
            'is_secure': parsed_url.scheme == 'https',
            'host': parsed_url.netloc,
            'path': parsed_url.path,
            'method': method.upper(),
            'queries': clean_queries(get_logged_params(url, params)),
            'slug': slug or self.slug,
            'request_timestamp': timezone.now(),
        }

        try:
            # Create the Request.
            req = Request(
                method=method.upper(), url=url, headers=headers, files=files, data=data or {}, json=json,
                params=params or {}, auth=auth, cookies=cookies, hooks=hooks,
            )
            prep = self.prepare_request(req)
            proxies = proxies or {}
            request_settings = self.merge_environment_settings(
                prep.url, proxies, stream, verify, cert
            )
            # Send the request.
            send_kwargs = {
                'timeout': timeout,
                'allow_redirects': allow_redirects,
            }
            send_kwargs.update(request_settings)
            logged_kwargs.update({
                'request_headers': clean_headers(dict(prep.headers.items())),
                'request_body': clean_body(prepare_body(prep.body), settings.LOG_REQUEST_BODY_LENGTH),
            })
            resp = self.send(prep, **send_kwargs)

            logged_kwargs.update({
                'response_timestamp': timezone.now(),
                'response_code': resp.status_code,
                'response_headers': clean_headers(dict(resp.headers.items())),
                'response_body': clean_body(prepare_body(resp.content), settings.LOG_RESPONSE_BODY_LENGTH),
                'status': LoggedRequest.get_status(resp.status_code)
            })
            return resp
        except Exception as ex:
            logged_kwargs.update({
                'error_description': force_text(ex),
                'status': LoggedRequest.CRITICAL,
                'exception_name': ex.__class__.__name__
            })
            raise
        finally:
            log_output_request(
                stringify_dict(logged_kwargs), self.related_objects if related_objects is None else related_objects
            )


def request(method, url, **kwargs):
    with SecuritySession() as session:
        return session.request(method=method, url=url, **kwargs)


def get(url, params=None, **kwargs):
    """Sends a GET request.
    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query string for the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    kwargs.setdefault('allow_redirects', True)
    return request('get', url, params=params, **kwargs)


def options(url, **kwargs):
    """Sends a OPTIONS request.
    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    kwargs.setdefault('allow_redirects', True)
    return request('options', url, **kwargs)


def head(url, **kwargs):
    """Sends a HEAD request.
    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    kwargs.setdefault('allow_redirects', False)
    return request('head', url, **kwargs)


def post(url, data=None, json=None, **kwargs):
    """Sends a POST request.
    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
    :param json: (optional) json data to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('post', url, data=data, json=json, **kwargs)


def put(url, data=None, **kwargs):
    """Sends a PUT request.
    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('put', url, data=data, **kwargs)


def patch(url, data=None, **kwargs):
    """Sends a PATCH request.
    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('patch', url, data=data, **kwargs)


def delete(url, **kwargs):
    """Sends a DELETE request.
    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('delete', url, **kwargs)



from datetime import timedelta

from ipware.ip import get_ip

from django.conf import settings
from django.utils import timezone
from django.utils.translation import ugettext as _

from .exception import ThrottlingException
from .models import InputLoggedRequest


class ThrottlingValidator:

    def __init__(self, timeframe, throttle_at, description):
        self.timeframe = timeframe
        self.throttle_at = throttle_at
        self.description = description

    def validate(self, request):
        if not getattr(settings, 'TURN_OFF_THROTTLING', False) and not self._validate(request):
            raise ThrottlingException(self.description)

    def _validate(self, request):
        raise NotImplemented

    def __repr__(self):
        return '<{} (timeframe={}, throttle_at={}, description={})>'.format(
            self.__class__.__name__, self.timeframe, self.throttle_at, self.description
        )


class PerRequestThrottlingValidator(ThrottlingValidator):

    def __init__(self, timeframe, throttle_at, description=_('Slow down')):
        super(PerRequestThrottlingValidator, self).__init__(timeframe, throttle_at, description)

    def _validate(self, request):
        count_same_requests = InputLoggedRequest.objects.filter(
            ip=get_ip(request), path=request.path,
            request_timestamp__gte=timezone.now() - timedelta(seconds=self.timeframe),
            method=request.method.upper()).count()
        return count_same_requests <= self.throttle_at


class LoginThrottlingValidator(ThrottlingValidator):

    def _validate(self, request):
        count_same_requests = InputLoggedRequest.objects.filter(
            ip=get_ip(request), path=request.path,
            request_timestamp__gte=timezone.now() - timedelta(seconds=self.timeframe),
            type=self.type).count()
        return count_same_requests <= self.throttle_at


class UnsuccessfulLoginThrottlingValidator(LoginThrottlingValidator):

    type = InputLoggedRequest.UNSUCCESSFUL_LOGIN_REQUEST

    def __init__(self, timeframe, throttle_at, description=_('Too many login attempts')):
        super().__init__(timeframe, throttle_at, description)


class SuccessfulLoginThrottlingValidator(LoginThrottlingValidator):

    type = InputLoggedRequest.SUCCESSFUL_LOGIN_REQUEST

    def __init__(self, timeframe, throttle_at, description=_('You have logged in too many times')):
        super().__init__(timeframe, throttle_at, description)


class UnSuccessfulTwoFactorCodeVerificationThrottlingValidator(LoginThrottlingValidator):

    type = InputLoggedRequest.UNSUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST

    def __init__(self, timeframe, throttle_at, description=_('Too many login attempts')):
        super().__init__(timeframe, throttle_at, description)


class SuccessfulTwoFactorCodeVerificationThrottlingValidator(LoginThrottlingValidator):

    type = InputLoggedRequest.SUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST

    def __init__(self, timeframe, throttle_at, description=_('You have logged in too many times')):
        super().__init__(timeframe, throttle_at, description)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-07-10 16:19
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0009_auto_20180418_0912'),
    ]

    operations = [
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='type',
            field=models.PositiveSmallIntegerField(
                choices=[(1, 'Common request'), (2, 'Throttled request'), (3, 'Successful login request'),
                         (4, 'Unsuccessful login request'), (5, 'Successful two factor code verification request'),
                         (6, 'Unsuccessful two factor code verification request')],
                default=1, verbose_name='type'),
        ),
    ]



import json

from django.core.serializers.json import DjangoJSONEncoder
from django.utils.html import format_html, mark_safe
from django.utils.translation import ugettext_lazy as _

from is_core.generic_views.inlines.inline_form_views import TabularInlineFormView
from is_core.main import UIRESTModelISCore
from is_core.utils.decorators import short_description

from security.models import CommandLog, InputLoggedRequest, OutputLoggedRequest, OutputLoggedRequestRelatedObjects

from ansi2html import Ansi2HTMLConverter


def display_json(value):
    return json.dumps(value, indent=4, ensure_ascii=False, cls=DjangoJSONEncoder)


def display_as_code(value):
    return format_html('<code style="white-space:pre-wrap;">{}</code>', value) if value else value


class RequestsLogISCore(UIRESTModelISCore):

    abstract = True

    can_create = can_update = can_delete = False

    @short_description(_('queries'))
    def queries_code(self, obj=None):
        return display_as_code(display_json(obj.queries)) if obj else None

    @short_description(_('request body'))
    def request_body_code(self, obj=None):
        return display_as_code(obj.request_body) if obj else None

    @short_description(_('request headers'))
    def request_headers_code(self, obj=None):
        return display_as_code(display_json(obj.request_headers)) if obj else None

    @short_description(_('response body'))
    def response_body_code(self, obj=None):
        return display_as_code(obj.response_body) if obj else None

    @short_description(_('response headers'))
    def response_headers_code(self, obj=None):
        return display_as_code(display_json(obj.response_headers)) if obj else None

    @short_description(_('error description'))
    def error_description_code(self, obj=None):
        return display_as_code(obj.error_description) if obj else None


class InputRequestsLogISCore(RequestsLogISCore):

    model = InputLoggedRequest
    ui_list_fields = (
        'request_timestamp', 'response_timestamp', 'response_time', 'status', 'response_code', 'host', 'short_path',
        'slug', 'ip', 'user', 'method', 'type', 'short_response_body', 'short_request_body', 'short_queries',
        'short_request_headers'
    )

    form_fieldsets = (
        (_('Request'), {'fields': ('request_timestamp', 'host', 'method', 'path', 'queries_code',
                                   'request_headers_code', 'request_body_code', 'is_secure')}),
        (_('Response'), {'fields': ('response_timestamp', 'response_code', 'status', 'response_headers_code',
                                    'response_body_code', 'type', 'error_description_code')}),
        (_('User information'), {'fields': ('user', 'ip')}),
        (_('Extra information'), {'fields': ('slug', 'response_time', 'output_logged_requests')}),
    )

    abstract = True


class OutputLoggedRequestRelatedObjectsInlineFormView(TabularInlineFormView):

    model = OutputLoggedRequestRelatedObjects
    fields = ('display_object',)


class OutputRequestsLogISCore(RequestsLogISCore):

    model = OutputLoggedRequest
    ui_list_fields = (
        'request_timestamp', 'response_timestamp', 'response_time', 'status', 'response_code', 'host', 'short_path',
        'method', 'slug', 'short_response_body', 'short_request_body', 'input_logged_request', 'short_queries',
        'short_request_headers'
    )

    form_fieldsets = (
        (_('Request'), {'fields': ('request_timestamp', 'host', 'method', 'path', 'queries_code',
                                   'request_headers_code', 'request_body_code', 'is_secure')}),
        (_('Response'), {'fields': ('response_timestamp', 'response_code', 'status', 'response_headers_code',
                                    'response_body_code', 'error_description_code')}),
        (_('Extra information'), {'fields': ('slug', 'response_time', 'input_logged_request')}),
        (_('Related objects'), {'inline_view': OutputLoggedRequestRelatedObjectsInlineFormView}),
    )

    abstract = True


class CommandLogISCore(UIRESTModelISCore):

    model = CommandLog

    create_permission = False
    update_permission = False
    delete_permission = False

    ui_list_fields = ('command_name', 'start', 'stop', 'executed_from_command_line', 'is_successful')

    form_fieldsets = (
        (None, {
            'fields': ('command_name', 'command_options', 'output_html'),
            'class': 'col-sm-6'
        }),
        (None, {
            'fields': ('start', 'stop', 'executed_from_command_line', 'is_successful'),
            'class': 'col-sm-6'
        }),
    )

    abstract = True

    @short_description(_('command output'))
    def output_html(self, obj=None):
        if obj and obj.output is not None:
            conv = Ansi2HTMLConverter()
            output = mark_safe(conv.convert(obj.output, full=False))
            return display_as_code(output)
        return None



import atexit
import gzip
import json
import os
import re
import sys
from datetime import datetime, time, timedelta
from importlib import import_module
from io import StringIO

from six.moves import input

from django.core import serializers
from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand
from django.core.serializers.json import DjangoJSONEncoder
from django.utils import timezone
from django.utils.encoding import force_bytes, force_text
from django.utils.timezone import now, utc
from django.urls import resolve
from django.urls.exceptions import Resolver404

from chamber.shortcuts import change_and_save

from .config import settings


UNIT_OPTIONS = {
    'h': lambda amount: timezone.now() - timedelta(hours=amount),
    'd': lambda amount: timezone.now() - timedelta(days=amount),
    'w': lambda amount: timezone.now() - timedelta(weeks=amount),
    'm': lambda amount: timezone.now() - timedelta(days=(30 * amount)),  # 30-day month
    'y': lambda amount: timezone.now() - timedelta(weeks=(52 * amount)),  # 364-day year
}


def is_base_collection(v):
    return isinstance(v, (list, tuple, set))


def get_throttling_validators(name):
    try:
        return getattr(import_module(settings.DEFAULT_THROTTLING_VALIDATORS_PATH), name)
    except (ImportError, AttributeError):
        raise ImproperlyConfigured('Throttling validator configuration {} is not defined'.format(name))


def get_headers(request):
    regex = re.compile('^HTTP_')
    return dict((regex.sub('', header), value) for (header, value)
                in request.META.items() if header.startswith('HTTP_'))


def remove_nul_from_string(value):
    return value.replace('\x00', '')


class TeeStringIO(StringIO):
    """
    StringIO that additionally writes to another stream(s).
    """
    ANSI_ESCAPE_REGEX = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')

    def __init__(self, streams):
        self.streams = streams
        super().__init__()

    def write(self, s):
        for stream in self.streams:
            if stream.isatty():
                stream.write(s)
            else:
                stream.write(self.ANSI_ESCAPE_REGEX.sub('', s))

        return super().write(s)

    def isatty(self):
        return True

    def flush(self):
        for stream in self.streams:
            stream.flush()


class CommandLogger(StringIO):
    """
    A helper class that runs a django command and logs details about its run into DB.
    """

    def __init__(self, command_function, stdout=None, stderr=None, **kwargs):
        """
        Initializes the command logger.

        Arguments:
            command_function: Callable that implements the command.
            stdout: Custom stream where command's standard output will be written.
            stderr: Custom stream where command's error output will be written.
            **kwargs: Keyword arguments passed to CommandLog model.
        """
        assert 'command_name' in kwargs, 'Key command_name must be provided in kwargs'

        self.command_function = command_function
        self.kwargs = kwargs
        self.stdout = stdout
        self.stderr = stderr

    def _start_intercepting(self):
        """
        Starts intercepting of stdout and stderr.
        """
        self.output = StringIO()
        self.output.isatty = lambda: True

        stdout_streams = [sys.__stdout__, self.output]
        if self.stdout:
            stdout_streams.append(self.stdout)
        sys.stdout = TeeStringIO(stdout_streams)

        stderr_streams = [sys.__stderr__, self.output]
        if self.stderr:
            stderr_streams.append(self.stderr)
        sys.stderr = TeeStringIO(stderr_streams)

    def _stop_intercepting(self):
        """
        Stops intercepting of stdout and stderr.
        """
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

    def run(self):
        """
        Runs the command function and returns its return value or re-raises any exceptions. The run of the command will
        not be logged if it is in excluded commands setting.
        """
        if self.kwargs['command_name'] in settings.COMMAND_LOG_EXCLUDED_COMMANDS:
            return self.command_function()

        self._start_intercepting()

        from security.models import CommandLog
        self.command_log = CommandLog(start=now(), **self.kwargs)
        self.command_log.save()

        # register call of the finish method in case the command exits the interpreter prematurely
        atexit.register(self._finish)

        try:
            ret_val = self.command_function()
            self._finish(success=True)
            return ret_val
        except Exception as ex:  # pylint: disable=W0703
            print(
                '{}: {}'.format(ex.__class__.__name__, ex),
                file=sys.stderr
            )
            self._finish(success=False)
            raise ex

    def _finish(self, success=False):
        if self.command_log.stop is None:
            change_and_save(
                self.command_log,
                output=self.output.getvalue(),
                stop=now(),
                is_successful=success
            )
            self._stop_intercepting()


class PurgeLogsBaseCommand(BaseCommand):

    help = ""
    args = '[amount duration]'

    model = None
    timestamp_field = None

    def get_timestamp_field(self):
        return self.timestamp_field

    def get_model(self):
        return self.model

    def add_arguments(self, parser):
        parser.add_argument('--expiration', action='store', dest='expiration',
                            help='Sets the timedelta from which logs will be removed.', required=True)
        parser.add_argument('--noinput', action='store_false', dest='interactive', default=True,
                            help='Tells Django to NOT prompt the user for input of any kind.')
        parser.add_argument('--backup', action='store', dest='backup', default=False,
                            help='Tells Django where to backup removed logs.')

    def serialize(self, qs):
        data = serializers.serialize('python', qs)

        for obj_data in data:
            del obj_data['pk']
        return data

    def backup_logs_and_clean_data(self, qs, backup_path):
        for timestamp in qs.datetimes(self.get_timestamp_field(), 'day'):
            min_timestamp = datetime.combine(timestamp, time.min).replace(tzinfo=utc)
            max_timestamp = datetime.combine(timestamp, time.max).replace(tzinfo=utc)
            qs_filtered_by_day = qs.filter(
                **{'{}__range'.format(self.get_timestamp_field()): (min_timestamp, max_timestamp)})

            if backup_path:
                log_file_path = os.path.abspath(os.path.join(backup_path, force_text(timestamp.date())))

                if os.path.isfile('{}.json.zip'.format(log_file_path)):
                    i = 0
                    while os.path.isfile('{}({}).json.zip'.format(log_file_path, i)):
                        i += 1
                    log_file_path = '{}({})'.format(log_file_path, i)

                self.stdout.write(4 * ' ' + 'generate backup file: ' + log_file_path)
                with gzip.open('{}.json.zip'.format(log_file_path), 'wb') as file_out:
                    file_out.write(
                        force_bytes(json.dumps(self.serialize(qs_filtered_by_day), cls=DjangoJSONEncoder, indent=5)))
            self.stdout.write(4 * ' ' + 'clean logs for day {}'.format(timestamp.date()))
            qs_filtered_by_day.delete()

    def handle(self, expiration, **options):
        # Check we have the correct values
        unit = expiration[-1]
        amount = expiration[0:-1]
        try:
            amount = int(amount)
        except ValueError:
            raise CommandError('Invalid expiration format')

        if unit not in UNIT_OPTIONS:
            raise CommandError('Invalid expiration format')

        model = self.get_model()
        qs = model.objects.filter(**{'{}__lte'.format(self.get_timestamp_field()): UNIT_OPTIONS[unit](amount)})

        if qs.count() == 0:
            self.stdout.write('There are no logs to delete.')
        else:
            if options.get('interactive'):
                confirm = input('''
                You have requested a database reset.
                This will IRREVERSIBLY DESTROY any
                logs created before {} {}
                ago. That is a total of {} logs.
                Are you sure you want to do this?
                Type 'yes' to continue, or 'no' to cancel: '''.format(amount, unit, qs.count()))
            else:
                confirm = 'yes'

            if confirm == 'yes':
                try:
                    self.stdout.write('Clean data')
                    self.backup_logs_and_clean_data(qs, options.get('backup'))
                except IOError as ex:
                    self.stderr.write(force_text(ex))


def regex_sub_groups_global(pattern, repl, string):
    """
    Globally replace all groups inside pattern with `repl`.
    If `pattern` doesn't have groups the whole match is replaced.
    """
    for search in reversed(list(re.finditer(pattern, string))):
        for i in range(len(search.groups()), 0 if search.groups() else -1, -1):
            start, end = search.span(i)
            string = string[:start] + repl + string[end:]
    return string


def get_view_from_request_or_none(request):
    try:
        return resolve(request.path).func
    except Resolver404:
        return None
